/*
 ============================================================================
 Name		 : Leaves.cpp
 Copyright   : Symbian Academy, 2007
 Description : Exercise: Leaves and the Cleanup Stack
 ============================================================================
 */

//  Include Files  

// !! -> Find Edit 1 in MainL() ->

#include <e32base.h>
#include <e32std.h>
#include <e32cons.h>			// Console

// Function prototype (C-style).
LOCAL_C void MainL();
LOCAL_C void DoLeaveTestsL();

//  Global Variables
LOCAL_D CConsoleBase* console;

// ------------------------------------------------------------------
// Definition of a sample C type class
// ------------------------------------------------------------------

// Edit 10: Remove the derivation of the CBase base class and execute
//          the application again.
//          Note that the destructor will no longer be executed by
//          the cleanup stack, as this is only automatically done
//          by the standard CleanupStack::PushL()-function - if the
//          object on the stack is derived from CBase.
//          Afterwards, add the CBase base class again.
class CLeaveClass : public CBase
	{
public:
	/** Constructor. */
	CLeaveClass();
	/** Destructor. */
	~CLeaveClass();
	/** This function intentionally causes a leave so that the effects
	 * can be tested. */
	void InitializeL();
	/** A simple status function that prints a line of text. */
	void PrintText();
	};

CLeaveClass::CLeaveClass()
	{
	_LIT(KConstructor, "Inside CLeaveClass::CLeaveClass()\n");
	console->Printf (KConstructor);
	}

CLeaveClass::~CLeaveClass()
	{
	_LIT(KDestructor, "Inside CLeaveClass::~CLeaveClass()\n");
	console->Printf (KDestructor);
	}

void CLeaveClass::InitializeL()
	{
	_LIT(KInitialize, "Inside CLeaveClass::InitializeL()\n");
	console->Printf (KInitialize);
	// Edit 7a: Intentionally cause a leave, e.g. with KErrGeneral.
	//          Its error code is -2.

	}

void CLeaveClass::PrintText()
	{
	_LIT(KPrintText, "Inside CLeaveClass::PrintText()\n");
	console->Printf (KPrintText);
	}

// ------------------------------------------------------------------
// Definition of a sample R type class
// ------------------------------------------------------------------
class RCleanupTest
	{
public:
	/** Constructor. */
	RCleanupTest();
	/** Here, the close function doesn't really free resources, but it
	 * is here to show how the Cleanup Stack calls this function automatically. */
	void Close();
	};

RCleanupTest::RCleanupTest()
	{
	_LIT(KRConstructor, "Inside RCleanupTest::RCleanupTest()\n");
	console->Printf (KRConstructor);
	}

void RCleanupTest::Close()
	{
	_LIT(KCloseText, "Inside RCleanupTest::Close()\n");
	console->Printf (KCloseText);
	}

// ------------------------------------------------------------------
// Main Application
// ------------------------------------------------------------------
GLDEF_C TInt E32Main()
	{
	// Create cleanup stack
	CTrapCleanup* cleanup = CTrapCleanup::New ();

	// Create output console
	_LIT(KTextConsoleTitle, "Console - Leaves");
	TRAPD(createError, console = Console::NewL(KTextConsoleTitle, TSize(KConsFullScreen,KConsFullScreen)))
	;
	if ( createError)
		return createError;

	// Run application code inside TRAP harness
	__UHEAP_MARK;

	TRAPD(mainError, MainL())
	;
	_LIT(KTextFailed, "MainL() failed, leave code = %d\n");
	if ( mainError)
		{
		console->Printf (KTextFailed, mainError);
		}

	// Wait for the space key before exiting.
	// Generally better than accepting any key to prevent immediate shutdown
	// if a key has still been in the queue when the app. started.
	_LIT(KTextPressSpace, "\n[Press space to exit]\n");
	console->Printf (KTextPressSpace);
	while (console->Getch ()!= EKeySpace)
		{
		}

	// Check the state of the heap memory - if the same amount of memory
	// is available as before MainL() was executed, everything's fine.
	// Otherwise, a panic will be raised, which would result in a graphical
	// error dialog in an UI framework, but exits the application when
	// using the console mode.
	__UHEAP_MARKEND;

	// If the program passed the __UHEAP_MARKEND-macro, no memory
	// leaks were detected
	_LIT(KTextMemoryLeak, "\nNo memory leaks detected! [Press space]\n");
	console->Printf (KTextMemoryLeak);
	while (console->Getch ()!= EKeySpace)
		{
		}

	// Cleanup
	delete console;
	delete cleanup;
	return KErrNone;
	}

LOCAL_C void MainL()
	{
	// ------------------------------------------------------------------
	// The Framework
	// ------------------------------------------------------------------
	// Edit 1: Uncomment the following line to see what it looks like if
	//         a leave gets through and is caught by the TRAPD-macro in E32Main().
	//         Afterwards, apply the comment again.
	// User::LeaveNoMemory();

	// Edit 2: Uncomment the following line to see what it looks like if
	//         a memory leak occurs in your console (!) application. 
	//         You will not get to the second "No memory leaks detected!"-message.
	//         Afterwards, apply the comment again.
	// TInt* x = new TInt;

	// -> Edit 3 -> go to the DoLeaveTestsL()-function below the MainL() function
	DoLeaveTestsL();

	// Edit 11: Use the TRAPD()-macro in the DoLeaveTestsL() function call in the
	//          line above to trap the leave.
	//          As you have already seen the error that is intentionally caused by
	//          the CLeaveClass::InitializeL()-function, we will ignore it now and
	//          continue with other tasks. As the leave will now be trapped right
	//          here, execution will continue in this function.
	//          However, remember that you should use TRAPs sparingly!

	// ------------------------------------------------------------------
	// Cleanup Stack for R classes
	// ------------------------------------------------------------------
	_LIT(KSeparator, "---------------------\n");
	console->Printf(KSeparator);

	// Edit 12: Take a look at the definition of RCleanupTest. It's a very simple
	//          class of the R type (not derived from CBase) and only has a
	//          Close()-method, which outputs a short status message.
	//
	//          Now, define a local, stack-based variable of type RCleanupTest
	//          with the name closeTest.


	// Edit 13: R classes do not have a destructor, instead they require Close()
	//          or Reset() to be called when the resource is no longer needed.
	//          To use the Cleanup Stack to call this function automatically,
	//          you have to push it using CleanupClosePushL().


	// Edit 15: Cause an intentional leave before the closeTest-object is destroyed
	//          through the Cleanup Stack function below.
	//          Note that Close() will still be executed, even though the 
	//          PopAndDestroy()-call will never be reached by the execution flow.


	// Edit 14: Clean the object as usual with CleanupStack::PopAndDestroy().
	//          As the closeTest-object was pushed differently, the Cleanup Stack
	//          knows that it should call closeTest->Close() instead of
	//          executing its destructor. Try if it works.

	}

LOCAL_C void DoLeaveTestsL()
	{
	// ------------------------------------------------------------------
	// Experiments with Leaves and the Cleanup Stack
	// ------------------------------------------------------------------

	// Note: We will try various tests in this function.
	//       Follow the instructions and their order (!!) closely, 
	//       as the edits will not be done strictly sequentially!
	//       -> Start with Edit 3!

	// Edit 6: Call the macro __UHEAP_FAILNEXT(1); before allocating
	//         the class. Because of the new (ELeave)-operator, this will
	//         cause a leave with the error code -4 (KErrNoMemory) and can be
	//         used to simulate a situation when allocation fails.
	//         Afterwards, apply the comment again.
	

	// Edit 3: Create an instance of the CLeaveClass on the heap using the
	//         special Symbian OS operator that leaves when there is a 
	//         problem with the allocation.
	//         Call the variable "test".
	

	// Edit 8: To prevent the memory leak caused by Edit 6,
	//         put the class on the Cleanup Stack.
	

	// -> Edit 7a -> Go to the InitializeL()-function of the CLeaveClass.

	// Edit 7b: Call the InitializeL()-function of the "test" instance.
	//         In its current implementation, this function will always
	//         cause a leave.
	//         Observe that the leave is caught by the TRAPD-macro in the
	//         E32Main()-function, but as the delete statement is never
	//         executed, our application now suffers from a memory leak.
	

	// Edit 4: Call the PrintText()-function of the "test"-object to see
	//         if the program reaches this line of code.
	

	// Edit 5: Delete test again using the conventional delete-function.
	//         Execute the application to see if the text is printed.
	//         No memory leak or error should occur.
	

	// Edit 9: Comment the delete-line that you added directly above for edit 4.
	//         Instead, use the Cleanup Stack to delete the object.
	//         Hint: CleanupStack::PopAndDestroy(test);
	//         Observe that the leave will of course still occur, but the
	//         memory leak is now gone.
	

	// -> Edit 10 -> Scroll up to the definition of CLeaveClass
	}
